
/* Necessary includes for device drivers */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h> /* printk() */
#include <linux/slab.h> /* kmalloc() */
#include <linux/fs.h> /* everything... */
#include <linux/cdev.h> /* everything... */
#include <linux/errno.h> /* error codes */
#include <linux/types.h> /* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h> /* O_ACCMODE */
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/list.h>
#include <linux/spinlock.h>
#include <asm/io.h>
#include <asm/system.h> /* cli(), *_flags */
#include <asm/uaccess.h> /* copy_from/to_user */
#include <asm/irq.h> 
#include <asm/irq_vectors.h>
#include <linux/mutex.h>
#include "public.h"

MODULE_LICENSE("Dual BSD/GPL");

static char *CANInfo = "$Id: CAN_LinuxDrv.ko 1.0 2013-10-28 9:18 mjm Exp $";

/* Declaration of CAN Drvr functions */
long candrv_ioctl(struct file *filp, unsigned int cmd, unsigned long data);
int candrv_open(struct inode *inode, struct file *filp);
int candrv_release(struct inode *inode, struct file *filp);
//ssize_t candrv_read(struct file *filp, char *buf, size_t count, loff_t *f_pos);
//ssize_t candrv_write(struct file *filp, char *buf, size_t count, loff_t *f_pos);
void candrv_exit(void);
int candrv_init(void);
static irqreturn_t candrv_intr(int irq, void *dev_id, struct pt_regs *regs);
long candrv_ioctl(struct file *filp, unsigned int cmd, unsigned long data);
void WRITE_PORT_UCHAR(unsigned char *port, unsigned char data);
char READ_PORT_UCHAR(unsigned char *port);


void write_command(unsigned char *port, unsigned char data);
char read_command(unsigned char *port);
/* Structure that declares the usual file */
/* access functions */
struct file_operations candrv_fops = {
  //read: candrv_read,
  //write: candrv_write,
  unlocked_ioctl: candrv_ioctl,
  open: candrv_open,
  release: candrv_release
};

/* Declaration of the init and exit functions */
module_init(candrv_init);
module_exit(candrv_exit);

/* Global variables of the driver */
/* Major number */
int candrv_major = 65;

//struct mutex pcan_fifo_prot;
struct mutex one_ioctl;


DRIVER_DATA	driverDat;
DRIVER_DATA *pDriver = &driverDat;


int candrv_init(void) {
  int result;

  /* Registering device */

  // Sign-on

  pDriver->numChannels = 2;

  printk("<100>WinSystems, Inc. Linux CAN Device Driver\n");

  printk("<100>Copyright 2013, All rights reserved\n");

  printk("<100>%s\n", CANInfo);

  result = register_chrdev(candrv_major, CANDRV_NAME, &candrv_fops);
  if (result < 0) {
    printk(
      "<100>candrv: cannot obtain major number %d\n", candrv_major);
    return result;
  }
  mutex_init(&one_ioctl);

  printk("<100>candrv:Inserting candrv module\n"); 
  return 0;

}


void candrv_exit(void)
{
	printk("<100>exit:Removing candrv module\n");
	disable_irq(pDriver->irq);
	printk("<100>exit:Freeing IRQ: %d \n", (int)pDriver->irq);
	free_irq(pDriver->irq, pDriver);
	release_region((unsigned long)pDriver->baseIO, (int)((pDriver->portOffset * pDriver->numChannels) - 1));
	unregister_chrdev(candrv_major, CANDRV_NAME);
}

int candrv_release(struct inode *inode, struct file *filp)
{
	disable_irq(pDriver->irq);
	printk("<100>release:Freeing IRQ: %d \n", (int)pDriver->irq);
	free_irq(pDriver->irq, pDriver);
	release_region((unsigned long)pDriver->baseIO, (int)((pDriver->portOffset * pDriver->numChannels) - 1));
	printk("<100>release:Removing candrv module\n");
  	/* Success */
  	return 0;
}



int candrv_open(struct inode *inode, struct file *filp)
{
 /* Success */
  return 0;
}

long candrv_ioctl(struct file *filp, unsigned int cmd, unsigned long data)
{

	INIT_DATA		init_cmd;
	READ_DATA		read_cmd;
	WRITE_DATA		write_cmd;
	NUM_PORTS		ports_cmd;
	VERSION_DATA	version_cmd;
	unsigned short  btr0btr1;
	struct resource *resource;
	int result;
	TPCANRdMsg   msg;
	//TPCANMsg *test;



	switch(cmd)
	{
	case INIT_CAN_DEVICE:

		copy_from_user((char *)&init_cmd, (char *)data, sizeof(INIT_DATA));

		pDriver->baseIO = init_cmd.baseIO;
		pDriver->bitRate = init_cmd.bitRate;
		pDriver->chan[0].PortBase = init_cmd.baseIO;
		pDriver->chan[1].PortBase = init_cmd.baseIO + init_cmd.portOffset;
		pDriver->portOffset = init_cmd.portOffset;
		pDriver->irq = init_cmd.irq;

		if(pDriver->numChannels > 2)
		{
			printk("<100>\ncandrv:Too many channels: %d\n", (int)pDriver->numChannels);
			return STATUS_DEVICE_CONFIGURATION_ERROR;
		}

		btr0btr1 = device_bitrate(pDriver->bitRate);
		//printk("<1>\ncandrv: Reached here -1\n");
		resource = request_region((long)pDriver->baseIO, (int)((pDriver->portOffset * pDriver->numChannels) - 1), CANDRV_NAME);
                printk("<100>\nrequest_region: base: 0x%08X  n: 0x%04X   Resources: 0x%08X  ", (int)pDriver->baseIO, (int)((pDriver->portOffset * pDriver->numChannels) - 1), (int)resource);
		if(resource) {

			if ((result = request_irq(pDriver->irq, (void *)candrv_intr, (int)NULL, CANDRV_NAME, (void *)pDriver))!=0){
			    printk("<100>\ncandrv:Failed request irq result = %d    \n", result);
			    release_region((unsigned long)pDriver->baseIO, (int)((pDriver->portOffset * pDriver->numChannels) - 1));
		            free_irq(pDriver->irq, pDriver);
			    return(STATUS_IRQ_REQUEST_ERROR);
			}
			else{
			    printk("<100>\ncandrv:Request irq ok\n");
			}
		}
		else {
			    printk("<100>\ncandrv:Failed IO region request\n");
			    release_region((unsigned long)pDriver->baseIO, (int)((pDriver->portOffset * pDriver->numChannels) - 1));
		            free_irq(pDriver->irq, pDriver);
			    return(STATUS_REQUEST_IO_REGION_ERROR);
		}


		//printk("<1>\ncandrv: Reached here 1\n");
		// switch to reset
		result = set_reset_mode(pDriver, 0);
		if (result != STATUS_SUCCESS) {
			printk("<100>\ncandrv:set_reset_mode channel 0 failed\n");
			release_region((unsigned long)pDriver->baseIO, (int)((pDriver->portOffset * pDriver->numChannels) - 1));
		        free_irq(pDriver->irq, pDriver);
			return STATUS_DEVICE_CONFIGURATION_ERROR;
		}
		//printk("<1>\ncandrv: Reached here 2\n");
		if(pDriver->numChannels > 1) {
	
			result = set_reset_mode(pDriver, 1);
			if (result != STATUS_SUCCESS) {
				printk("<100>\ncandrv:set_reset_mode channel 1 failed\n");
				release_region((unsigned long)pDriver->baseIO, (int)((pDriver->portOffset * pDriver->numChannels) - 1));
			        free_irq(pDriver->irq, pDriver);
				return STATUS_DEVICE_CONFIGURATION_ERROR;
			}
		}

		//printk("<1>\ncandrv: Reached here 3\n");
		// configure clock divider register, switch into pelican mode, depended of of type
		WRITE_PORT_UCHAR(pDriver->chan[0].PortBase + CLKDIVIDER, BASICCAN_DEFAULT);
		

		// configure acceptance code register
		WRITE_PORT_UCHAR(pDriver->chan[0].PortBase + ACCEPTANCE_CODE, 0);


		// configure acceptance mask register to don't care
		WRITE_PORT_UCHAR(pDriver->chan[0].PortBase + ACCEPTANCE_MASK, 0xff);

		// configure bus timing registers
		WRITE_PORT_UCHAR(pDriver->chan[0].PortBase + TIMING0, (unsigned char)((btr0btr1 >> 8) & 0xff));
		WRITE_PORT_UCHAR(pDriver->chan[0].PortBase + TIMING1, (unsigned char)((btr0btr1     ) & 0xff));

		// configure output control registers
		WRITE_PORT_UCHAR(pDriver->chan[0].PortBase + OUTPUT_CONTROL, OUTPUT_CONTROL_SETUP);


		// clear any pending interrupt
		READ_PORT_UCHAR( pDriver->chan[0].PortBase + INTERRUPT_STATUS );
		//printk("<1>\ncandrv: Reached here 4\n");
		if(pDriver->numChannels > 1) {


			// configure clock divider register, switch into pelican mode, depended of of type
			WRITE_PORT_UCHAR(pDriver->chan[1].PortBase + CLKDIVIDER, BASICCAN_DEFAULT);
		

			// configure acceptance code register
			WRITE_PORT_UCHAR(pDriver->chan[1].PortBase + ACCEPTANCE_CODE, 0);


			// configure acceptance mask register to don't care
			WRITE_PORT_UCHAR(pDriver->chan[1].PortBase + ACCEPTANCE_MASK, 0xff);

			// configure bus timing registers
			WRITE_PORT_UCHAR(pDriver->chan[1].PortBase + TIMING0, (unsigned char)((btr0btr1 >> 8) & 0xff));
			WRITE_PORT_UCHAR(pDriver->chan[1].PortBase + TIMING1, (unsigned char)((btr0btr1     ) & 0xff));

			// configure output control registers
			WRITE_PORT_UCHAR(pDriver->chan[1].PortBase + OUTPUT_CONTROL, OUTPUT_CONTROL_SETUP);


			// clear any pending interrupt
			READ_PORT_UCHAR(pDriver->chan[1].PortBase + INTERRUPT_STATUS );
		}
		//printk("<1>\ncandrv: Reached here 5\n");
		// enter normal operating mode
		result = set_normal_mode(pDriver, 0);
		if (result) {
			printk("<100>\ncandrv:set_normal_mode channel 0 failed\n");
			release_region((unsigned long)pDriver->baseIO, (int)((pDriver->portOffset * 2) - 1));
		        free_irq(pDriver->irq, pDriver);
			return STATUS_DEVICE_CONFIGURATION_ERROR;
		}
		//printk("<1>\ncandrv: Reached here 6\n");
		if(pDriver->numChannels > 1) {

			result = set_normal_mode(pDriver, 1);
			if (result) {
				printk("<100>\ncandrv:set_normal_mode channel 1 failed\n");
				release_region((unsigned long)pDriver->baseIO, (int)((pDriver->portOffset * 2) - 1));
			        free_irq(pDriver->irq, pDriver);
				return STATUS_DEVICE_CONFIGURATION_ERROR;
			}
		}

		//printk("<1>\ncandrv: Reached here 7\n");
		pcan_fifo_init(&pDriver->chan[0].readFifo,   &pDriver->chan[0].rMsg[0], &pDriver->chan[0].rMsg[READ_MESSAGE_COUNT - 1],  READ_MESSAGE_COUNT,  sizeof(TPCANRdMsg));
		pcan_fifo_init(&pDriver->chan[0].writeFifo,  &pDriver->chan[0].wMsg[0], &pDriver->chan[0].wMsg[WRITE_MESSAGE_COUNT - 1], WRITE_MESSAGE_COUNT, sizeof(TPCANMsg) );

		pcan_fifo_init(&pDriver->chan[1].readFifo,   &pDriver->chan[1].rMsg[0], &pDriver->chan[1].rMsg[READ_MESSAGE_COUNT - 1],  READ_MESSAGE_COUNT,  sizeof(TPCANRdMsg));
		pcan_fifo_init(&pDriver->chan[1].writeFifo,  &pDriver->chan[1].wMsg[0], &pDriver->chan[1].wMsg[WRITE_MESSAGE_COUNT - 1], WRITE_MESSAGE_COUNT, sizeof(TPCANMsg) );



		//printk("<1>\ncandrv: Reached here 8\n");
		// enable CAN interrupts
		enable_irq(pDriver->irq);
		device_irq_enable(pDriver);
		printk("<100>\ncandrv:Init complete status: success!\n");

		result = STATUS_SUCCESS;


		break;


    case IOCTL_GET_VERSION:          // code == 0x800


 
		version_cmd.version = 0x00010000;
		copy_to_user((char *)data, &version_cmd, sizeof(VERSION_DATA));

        result = STATUS_SUCCESS;
        break;

    case IOCTL_CANDRV_READ:          // code == 0x800

		mutex_lock(&one_ioctl);
		disable_irq(pDriver->irq);

		copy_from_user((char *)&read_cmd, (char *)data, sizeof(READ_DATA));

		
		//device_read_frames(pDriver, read_cmd.channel);


		result = pcan_fifo_get(&pDriver->chan[read_cmd.channel].readFifo, &msg);
		if(result == 0) {
			copy_to_user(read_cmd.message, &msg, sizeof(TPCANRdMsg));
        		result = STATUS_SUCCESS;
		}
		else {
			result = STATUS_NO_DATAGRAM_PRESENT;
		}

		enable_irq(pDriver->irq);
		mutex_unlock(&one_ioctl);


		break;

    case IOCTL_CANDRV_NUMPORT:          


		copy_from_user((char *)&ports_cmd, (char *)data, sizeof(NUM_PORTS));

		pDriver->numChannels = ports_cmd.numPorts;
		printk("<100>\ncandrv: numPorts set at: %d\n", (int)pDriver->numChannels);
		result = STATUS_SUCCESS;


		break;


    case IOCTL_CANDRV_WRITE:          // code == 0x800


		mutex_lock(&one_ioctl);
		disable_irq(pDriver->irq);

		copy_from_user((char *)&write_cmd, (char *)data, sizeof(WRITE_DATA));
		//test = (TPCANMsg *)write_cmd.message;
		//printk("<100>\nMessage ID: 0x%X Data[0]: %x  Data[1]; %x \n", (unsigned int)test->ID, test->DATA[0], test->DATA[1]);


		result = pcan_fifo_put(&pDriver->chan[write_cmd.channel].writeFifo, (void *)write_cmd.message);


		if(result) {
			result = STATUS_INSUFFICIENT_RESOURCES;

			//status = STATUS_SUCCESS;
		}	
		else
		{

			device_write(pDriver, write_cmd.channel);       
        		result = STATUS_SUCCESS;
		}	

		enable_irq(pDriver->irq);
		mutex_unlock(&one_ioctl);

		break;


    default:
        result = STATUS_INVALID_DEVICE_REQUEST;
        break;
    }


	return result;
}

//----------------------------------------------------------------------------
// switches the chip into reset mode
int set_reset_mode(DRIVER_DATA *pDriver, unsigned short channel)
{
	unsigned long time;
	unsigned char  tmp;


	
	tmp = READ_PORT_UCHAR(pDriver->chan[channel].PortBase + CONTROL );

	for(time = 0x100; (time != 0) && !(tmp & RESET_MODE); time--)
	{
		WRITE_PORT_UCHAR(pDriver->chan[channel].PortBase + CONTROL, RESET_MODE);
		WRITE_PORT_UCHAR((unsigned char *)0x080, 0);
		//udelay(1);
	    //delay.QuadPart = WDF_REL_TIMEOUT_IN_MS(10);
	//	KeDelayExecutionThread(KernelMode, FALSE, &delay);


		tmp = READ_PORT_UCHAR(pDriver->chan[channel].PortBase + CONTROL );
	}

	if (!(tmp & RESET_MODE))
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	else
		return STATUS_SUCCESS;
}

//----------------------------------------------------------------------------
// switches the chip back from reset mode
int set_normal_mode(DRIVER_DATA *pDriver, unsigned short channel)
{
	//u32 dwStart = get_mtime();
	unsigned char tmp;
	unsigned long time;


	tmp = READ_PORT_UCHAR(pDriver->chan[channel].PortBase + CONTROL );
	
	for(time = 0x100; (tmp & RESET_MODE) && (time != 0); time--)
	{
		WRITE_PORT_UCHAR(pDriver->chan[channel].PortBase + CONTROL, tmp & ~RESET_MODE);
		WRITE_PORT_UCHAR((unsigned char *)0x080, 0);
		//udelay(1);
	    	//delay.QuadPart = WDF_REL_TIMEOUT_IN_MS(10);
		//KeDelayExecutionThread(KernelMode, FALSE, &delay);
		tmp = READ_PORT_UCHAR(pDriver->chan[channel].PortBase + CONTROL );
	}

	if (tmp & RESET_MODE)
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	else
		return STATUS_SUCCESS;
}


//----------------------------------------------------------------------------
// interrupt enable and disable
//void sja1000_irq_enable_mask(PWCANDRV_DEVICE_CONTEXT   devExt, UCHAR mask)
//{
//	WRITE_PORT_UCHAR(devExt->PortBase + INTERRUPT_ENABLE, mask);
//}
//
//void sja1000_irq_disable_mask(PWCANDRV_DEVICE_CONTEXT   devExt, UCHAR mask)
//{
//	WRITE_PORT_UCHAR(devExt->PortBase + INTERRUPT_ENABLE, ~mask);
//}

void device_irq_enable(DRIVER_DATA *pDriver)
{
	unsigned char tmp;

	//tmp = READ_PORT_UCHAR(pDriver->chan[0].PortBase + CONTROL);
	//tmp = (TRANSMIT_INTERRUPT_ENABLE | RECEIVE_INTERRUPT_ENABLE);
	tmp = (TRANSMIT_INTERRUPT_ENABLE | RECEIVE_INTERRUPT_ENABLE | DATA_OVERRUN_INTERRUPT_ENABLE | BUS_ERROR_INTERRUPT_ENABLE | ERROR_PASSIV_INTERRUPT);
	WRITE_PORT_UCHAR(pDriver->chan[0].PortBase + CONTROL, tmp);

	printk("<100>\nEnable channel 0 at address 0x%08X ", (int)(pDriver->chan[0].PortBase + CONTROL));
	printk("<100>\nEnable value written: 0x%X  ", tmp);

	if(pDriver->numChannels > 1) {

		//tmp = READ_PORT_UCHAR(pDriver->chan[1].PortBase + CONTROL);
		//tmp = (TRANSMIT_INTERRUPT_ENABLE | RECEIVE_INTERRUPT_ENABLE);
		tmp = (TRANSMIT_INTERRUPT_ENABLE | RECEIVE_INTERRUPT_ENABLE | DATA_OVERRUN_INTERRUPT_ENABLE | BUS_ERROR_INTERRUPT_ENABLE | ERROR_PASSIV_INTERRUPT);
		WRITE_PORT_UCHAR(pDriver->chan[1].PortBase + CONTROL, tmp);

		printk("<100>\nEnable channel 1 at address 0x%08X ", (int)(pDriver->chan[1].PortBase + CONTROL));
		printk("<100>\nEnable value written: 0x%X  ", tmp);
	}



}

void device_irq_disable(DRIVER_DATA *pDriver)
{
	unsigned char tmp;

	tmp = READ_PORT_UCHAR(pDriver->chan[0].PortBase + CONTROL);
	//tmp &= ~(TRANSMIT_INTERRUPT_ENABLE | RECEIVE_INTERRUPT_ENABLE);
	tmp &= ~(TRANSMIT_INTERRUPT_ENABLE | RECEIVE_INTERRUPT_ENABLE | DATA_OVERRUN_INTERRUPT_ENABLE | BUS_ERROR_INTERRUPT_ENABLE | ERROR_PASSIV_INTERRUPT);
	WRITE_PORT_UCHAR(pDriver->chan[0].PortBase + CONTROL, tmp);

	if(pDriver->numChannels > 1) {

		tmp = READ_PORT_UCHAR(pDriver->chan[1].PortBase + CONTROL);
		//tmp &= ~(TRANSMIT_INTERRUPT_ENABLE | RECEIVE_INTERRUPT_ENABLE);
		tmp &= ~(TRANSMIT_INTERRUPT_ENABLE | RECEIVE_INTERRUPT_ENABLE | DATA_OVERRUN_INTERRUPT_ENABLE | BUS_ERROR_INTERRUPT_ENABLE | ERROR_PASSIV_INTERRUPT);
		WRITE_PORT_UCHAR(pDriver->chan[1].PortBase + CONTROL, tmp);
	}
}

//----------------------------------------------------------------------------



//----------------------------------------------------------------------------
// read CAN-data from chip, supposed a message is available
int device_read_frames(DRIVER_DATA *pDriver, unsigned short channel)
{
	int msgs   = MAX_MESSAGES_PER_INTERRUPT;
	unsigned char fi;
	unsigned char dreg;
	unsigned char dlc;
	ULCONV localID;
	struct can_frame frame;
	int i;
	int result = 0;


	if(!(READ_PORT_UCHAR(pDriver->chan[channel].PortBase + CHIPSTATUS) & RECEIVE_BUFFER_STATUS))
		return result;


	do
	{
		if(!(READ_PORT_UCHAR(pDriver->chan[channel].PortBase + CHIPSTATUS) & RECEIVE_BUFFER_STATUS))
			break;
	
		fi  = READ_PORT_UCHAR(pDriver->chan[channel].PortBase + RECEIVE_FRAME_BASE + 1);
		dlc = fi & BUFFER_DLC_MASK;

		if (dlc > 8)
			dlc = 8;


		/* standard frame format  */
		dreg = RECEIVE_FRAME_BASE + 2;

		localID.ul    = 0;

		localID.uc[1] = READ_PORT_UCHAR(pDriver->chan[channel].PortBase + RECEIVE_FRAME_BASE + 0);
		localID.uc[0] = READ_PORT_UCHAR(pDriver->chan[channel].PortBase + RECEIVE_FRAME_BASE + 1);

		frame.can_id = localID.ul;



		if (fi & BUFFER_RTR)
			frame.can_id |= CAN_RTR_FLAG;

		for (i = 0; i < dlc; i++) {
			frame.data[i] = READ_PORT_UCHAR(pDriver->chan[channel].PortBase + dreg++);
		}

		frame.can_id |= dlc;

		//SJA1000_LOCK_IRQSAVE(in_lock);

		if ((i = pcan_chardev_rx(pDriver, &frame, channel))) // put into specific data sink
			result = i; // save the last result

		// release the receive buffer
		guarded_write_command(pDriver, RELEASE_RECEIVE_BUFFER, channel);

		// give time to settle
		//udelay(1);

	} while (READ_PORT_UCHAR(pDriver->chan[channel].PortBase + CHIPSTATUS) & RECEIVE_BUFFER_STATUS && (msgs--));

	return result;
}


//----------------------------------------------------------------------------
// write CAN-data to chip,
int device_write_frame(DRIVER_DATA *pDriver, struct can_frame *cf, unsigned short channel)
{
	unsigned char dlc;
	ULCONV localID;
	unsigned char dreg;
	int i;

	dlc = (unsigned char)(cf->can_id & BUFFER_DLC_MASK);
	localID.ul = cf->can_id;


	dreg = TRANSMIT_FRAME_BASE + 2;

	WRITE_PORT_UCHAR(pDriver->chan[channel].PortBase + TRANSMIT_FRAME_BASE , localID.uc[1]);
	WRITE_PORT_UCHAR(pDriver->chan[channel].PortBase + TRANSMIT_FRAME_BASE + 1, localID.uc[0]);


	for (i = 0; i < dlc; i++) 
	{
		WRITE_PORT_UCHAR(pDriver->chan[channel].PortBase + dreg++, cf->data[i]);
	}

	// request a transmission
	guarded_write_command(pDriver, TRANSMISSION_REQUEST, channel);

	return 0;
}

/* Called by isr */
int isr_device_write(DRIVER_DATA *pDriver, unsigned short channel)
{
	int        result;
	TPCANMsg   msg;
	struct can_frame cf;


	// get a fifo element and step forward
	result = pcan_fifo_get(&pDriver->chan[channel].writeFifo, &msg);
	if (result)
		return result;

	/* convert from old style FIFO message until FIFO supports new */
	/* struct can_frame and error frames */
	msg2frame(&cf, &msg);

	device_write_frame(pDriver, &cf, channel);

	return 0;
}


//----------------------------------------------------------------------------
// write CAN-data from FIFO to chip
int device_write(DRIVER_DATA *pDriver, unsigned short channel)
{
	int err = 0;


	/* Check if Tx buffer is empty before writing on */
	if (READ_PORT_UCHAR(pDriver->chan[channel].PortBase + CHIPSTATUS) & TRANS_BUFFER_STATUS)
	{

		err = isr_device_write(pDriver, channel);

	}

	return err;
}

//----------------------------------------------------------------------------
// handle a interrupt request
static irqreturn_t candrv_intr(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned char irqstatus;
	unsigned char rxBufferStatus;
	int err;
	int j = MAX_INTERRUPTS_PER_ENTRY;
	unsigned short channel = 0;
	DRIVER_DATA *driverPtr;
	unsigned char chipstatus;
	struct can_frame ef;

	driverPtr = (DRIVER_DATA *)dev_id;

	device_irq_disable(driverPtr);
	printk("<100>\nEntered Interrupt Handler ");

	while(channel < 2) {
		while (j-- && ((irqstatus = (READ_PORT_UCHAR(driverPtr->chan[channel].PortBase + INTERRUPT_STATUS) & INTERRUPT_MASK)) || (rxBufferStatus = (READ_PORT_UCHAR(driverPtr->chan[channel].PortBase + CHIPSTATUS) & 0x01))))
		{


			if (irqstatus & TRANSMIT_INTERRUPT)
			{
				printk("<100>\nTransmit interrupt channel: %d ", channel);
				// handle transmission
				device_write(driverPtr, channel);
			}


			if ((irqstatus & RECEIVE_INTERRUPT) || (READ_PORT_UCHAR(driverPtr->chan[channel].PortBase + CHIPSTATUS) & 0x01))
			{

				printk("<100>\nReceive interrupt channel: %d ", channel);
				// handle receiption
				if ((err = device_read_frames(driverPtr, channel)) <= 0) /* put to input queues */
				{
					// throw away last message which was refused by fifo
					guarded_write_command(driverPtr, RELEASE_RECEIVE_BUFFER, channel);
				}
			}

			if(irqstatus & DATA_OVERRUN_INTERRUPT)
			{
				WRITE_PORT_UCHAR(driverPtr->chan[channel].PortBase + COMMAND, CLEAR_DATA_OVERRUN);
			}

			chipstatus = READ_PORT_UCHAR(driverPtr->chan[channel].PortBase + CHIPSTATUS);

	
			if (irqstatus & (BUS_ERROR_INTERRUPT | ERROR_PASSIV_INTERRUPT | ERROR_WARN_INTERRUPT))
			{

				if (irqstatus & (ERROR_PASSIV_INTERRUPT | ERROR_WARN_INTERRUPT))
				{

					switch (chipstatus & (BUS_STATUS | ERROR_STATUS))
					{

					case BUS_STATUS:
						// bus-off
						set_normal_mode(driverPtr, channel);
						break;
					case ERROR_STATUS:
						if (irqstatus & ERROR_PASSIV_INTERRUPT)
						{
							// either enter or leave error passive status

							ef.can_id  |= CAN_ERR_CRTL;
							ef.data[1] |= (CAN_ERR_CRTL_RX_PASSIVE | CAN_ERR_CRTL_TX_PASSIVE);

						
						}
						else
						{
							// it was a warning limit reached event

							ef.can_id  |= CAN_ERR_CRTL;
							ef.data[1] |= (CAN_ERR_CRTL_RX_WARNING | CAN_ERR_CRTL_TX_WARNING);
						}
						break;
					}
				}
			}



		}
		channel++;
		if((driverPtr->numChannels < 2) && (channel >= 1)) 
			break;
	}

	device_irq_enable(driverPtr);

	//irqstatus = READ_PORT_UCHAR(driverPtr->chan[0].PortBase + INTERRUPT_STATUS);

	//if(pDriver->numChannels > 1) 
	//	irqstatus = READ_PORT_UCHAR(driverPtr->chan[1].PortBase + INTERRUPT_STATUS);

	return IRQ_HANDLED;

}




//----------------------------------------------------------------------------
// calculate BTR0BTR1 for odd bitrates
//
// most parts of this code is from Arnaud Westenberg email:arnaud@wanadoo.nl
// www.home.wanadoo.nl/arnaud
//
// Set communication parameters.
// param rate baud rate in Hz
// param clock frequency of sja1000 clock in Hz
// param sjw synchronization jump width (0-3) prescaled clock cycles
// param sampl_pt sample point in % (0-100) sets (TSEG1+2)/(TSEG1+TSEG2+3) ratio
// param flags fields BTR1_SAM, OCMODE, OCPOL, OCTP, OCTN, CLK_OFF, CBP
//
int device_baud_rate(int rate, int flags)
{
	int best_error = 1000000000;
	int error;
	int best_tseg=0, best_brp=0, best_rate=0, brp=0;
	int tseg=0, tseg1=0, tseg2=0;
	int clock = CLOCK_HZ / 2;
	unsigned short wBTR0BTR1;
	int sjw = 0;
	int sampl_pt = 90;

	// some heuristic specials
	if (rate > ((1000000 + 500000) / 2))
		sampl_pt = 75;

	if (rate < ((12500 + 10000) / 2))
		sampl_pt = 75;

	if (rate < ((100000 + 125000) / 2))
		sjw = 1;

	// tseg even = round down, odd = round up
	for (tseg = (0 + 0 + 2) * 2; tseg <= (MAX_TSEG2 + MAX_TSEG1 + 2) * 2 + 1; tseg++)
	{
		brp = clock / ((1 + tseg / 2) * rate) + tseg % 2;
		if ((brp == 0) || (brp > 64))
			continue;

		error = rate - clock / (brp * (1 + tseg / 2));
		if (error < 0)
			error = -error;

		if (error <= best_error)
		{
			best_error = error;
			best_tseg = tseg/2;
			best_brp = brp-1;
			best_rate = clock/(brp*(1+tseg/2));
		}
	}

	if (best_error && (rate / best_error < 10))
	{

		return 0;
	}

	tseg2 = best_tseg - (sampl_pt * (best_tseg + 1)) / 100;

	if (tseg2 < 0)
		tseg2 = 0;

	if (tseg2 > MAX_TSEG2)
		tseg2 = MAX_TSEG2;

	tseg1 = best_tseg - tseg2 - 2;

	if (tseg1 > MAX_TSEG1)
	{
		tseg1 = MAX_TSEG1;
		tseg2 = best_tseg-tseg1-2;
	}

	wBTR0BTR1 = ((sjw<<6 | best_brp) << 8) | (((flags & BTR1_SAM) != 0)<<7 | tseg2<<4 | tseg1);

	return wBTR0BTR1;
}

//----------------------------------------------------------------------------
// get BTR0BTR1 init values
unsigned short device_bitrate(unsigned long dwBitRate)
{
	unsigned short wBTR0BTR1 = 0;;

	// get default const values
	switch (dwBitRate)
	{
	case 1000000: wBTR0BTR1 = CAN_BAUD_1M;   break;
	case  500000: wBTR0BTR1 = CAN_BAUD_500K; break;
	case  250000: wBTR0BTR1 = CAN_BAUD_250K; break;
	case  125000: wBTR0BTR1 = CAN_BAUD_125K; break;
	case  100000: wBTR0BTR1 = CAN_BAUD_100K; break;
	case   50000: wBTR0BTR1 = CAN_BAUD_50K;  break;
	case   20000: wBTR0BTR1 = CAN_BAUD_20K;  break;
	case   10000: wBTR0BTR1 = CAN_BAUD_10K;  break;
	case    5000: wBTR0BTR1 = CAN_BAUD_5K;   break;
	case       0: wBTR0BTR1 = 0;             break;

	default:
		 // calculate for exotic values
		 wBTR0BTR1 = (unsigned short)device_baud_rate((unsigned short)dwBitRate, 0);
	}

	return wBTR0BTR1;
}

// maybe this goes to a new file pcan_chardev.c some day.
int pcan_chardev_rx(DRIVER_DATA *pDriver, struct can_frame *cf, unsigned short channel)
{
  int result = 0;

 
  TPCANRdMsg msg;

  /* convert to old style FIFO message until FIFO supports new */
  /* struct can_frame and error frames */
  frame2msg(cf, &msg.Msg);

  // step forward in fifo
  result = pcan_fifo_put(&pDriver->chan[channel].readFifo, &msg);

  // flag to higher layers that a message was put into fifo or an error occurred
  result = (result == -1) ? result : 1;
 

  return result;
}

//----------------------------------------------------------------------------
// guards writing sja1000's command register in multicore environments
void guarded_write_command(DRIVER_DATA *pDriver, unsigned char data, unsigned short channel)
{

	WRITE_PORT_UCHAR(pDriver->chan[channel].PortBase + COMMAND, data);
	READ_PORT_UCHAR(pDriver->chan[channel].PortBase + CHIPSTATUS); // draw a breath after writing the command register

}

//----------------------------------------------------------------------------
// convert struct can_frame to struct TPCANMsg
// To reduce the complexity (and CPU usage) there are no checks (e.g. for dlc)
// here as it is assumed that the creator of the source struct has done this work
void frame2msg(struct can_frame *cf, TPCANMsg *msg)
{

  if (cf->can_id & BUFFER_RTR)
    msg->MSGTYPE = MSGTYPE_RTR;
  else
    msg->MSGTYPE = MSGTYPE_STANDARD;

  msg->ID  = cf->can_id & 0xffD0; 
  msg->ID >>= 5;
  msg->LEN = cf->can_id & 0x000f; /* no need to check value range here */

  memcpy(&msg->DATA[0], &cf->data[0], 8); /* also copy trailing zeros */
}

//----------------------------------------------------------------------------
// convert struct TPCANMsg to struct can_frame
// To reduce the complexity (and CPU usage) there are no checks (e.g. for dlc)
// here as it is assumed that the creator of the source struct has done this work
void msg2frame(struct can_frame *cf, TPCANMsg *msg)
{
  cf->can_id = (unsigned short)msg->ID;

  cf->can_id <<= 5;

  if (msg->MSGTYPE & MSGTYPE_RTR)
    cf->can_id |= BUFFER_RTR;


  // if (msg->MSGTYPE & MSGTYPE_??????)
  //   cf->can_id |= CAN_ERR_FLAG;

  cf->can_id |= (msg->LEN & 0x000f); /* no need to check value range here */

  memcpy(&cf->data[0], &msg->DATA[0], 8); /* also copy trailing zeros */
}


int pcan_fifo_get(FIFO_MANAGER *anchor, void *pvGetData)
{
  int err = 0;
  unsigned long temp;

  ///mutex_lock(&pcan_fifo_prot);

  if (anchor->nStored > 0)
  {
    memcpy(pvGetData, anchor->r, anchor->wCopySize);

    anchor->nStored--;
    temp = (unsigned long)anchor->r;
    if (anchor->r < anchor->bufferEnd)
      temp += (unsigned long)anchor->wStepSize;  // increment to next
    else
      temp = (unsigned long)anchor->bufferBegin; // start from begin
    anchor->r = (void *)temp;
  }
  else
    err = -1; //-ENODATA;

  //mutex_unlock(&pcan_fifo_prot);

  return err;
}

//----------------------------------------------------------------------------
// find the proper clock divider
//UCHAR clkdivider(DRIVER_DATA *pDriver)
//{
//	if (!devExt->props.ucExternalClock)  // crystal based
//		return BASICCAN_DEFAULT;
//
//	// configure clock divider register, switch into pelican mode, depended of of type
//	switch (devExt->props.ucMasterDevice)
//	{
//	case CHANNEL_SLAVE:
//	case CHANNEL_SINGLE:
//		return BASICCAN_SINGLE;  // neither a slave nor a single device distribute the clock ...
//
//	default:
//		return BASICCAN_MASTER;  // ... but a master does
//	}
//}

int pcan_fifo_put(FIFO_MANAGER *anchor, void *pvPutData)
{
  int err = 0;
  unsigned long temp;

  //mutex_lock(&pcan_fifo_prot);

  if (anchor->nStored < anchor->nCount)
  {
    memcpy(anchor->w, pvPutData, anchor->wCopySize);

    anchor->nStored++;
    anchor->dwTotal++;

    temp = (unsigned long)anchor->w;
    if (anchor->w < anchor->bufferEnd)
      temp += (unsigned long)anchor->wStepSize;   // increment to next
    else
      temp = (unsigned long)anchor->bufferBegin;  // start from begin
    anchor->w = (void *)temp;
  }
  else
    err = -1; //-ENOSPC;

  //mutex_unlock(&pcan_fifo_prot);

  return err;
}

int pcan_fifo_init(FIFO_MANAGER *anchor, void *bufferBegin, void *bufferEnd, int nCount, unsigned short wCopySize)
{
  anchor->wCopySize   = wCopySize;
  anchor->wStepSize   = ((unsigned long)bufferBegin == (unsigned long)bufferEnd) ? (unsigned short)0 : (unsigned short)(((unsigned long)bufferEnd - (unsigned long)bufferBegin) / (unsigned long)(nCount - 1));
  anchor->nCount      = nCount;
  anchor->bufferBegin = bufferBegin;
  anchor->bufferEnd   = bufferEnd;

  // check for fatal program errors
  if ((anchor->wStepSize < anchor->wCopySize) || (anchor->bufferBegin > anchor->bufferEnd) || (nCount <= 1))
    return -1; //EINVAL;

  return pcan_fifo_reset(anchor);
}

int pcan_fifo_reset(FIFO_MANAGER *anchor)
{

  anchor->dwTotal       = 0;
  anchor->nStored       = 0;
  anchor->r = anchor->w = anchor->bufferBegin; // nothing to read

  return 0;
}

void WRITE_PORT_UCHAR(unsigned char *port, unsigned char data)
{

	unsigned char j;

	outb(data, (unsigned long)port);
	j = inb((unsigned long)port);
}

char READ_PORT_UCHAR(unsigned char *port)
{

	unsigned char j;

	j = inb((unsigned long)port);
	return(j);
}








