
//

#ifndef PUBLIC_H
#define PUBLIC_H

#define IOCTL_GET_VERSION        101
#define IOCTL_CANDRV_READ        102
#define IOCTL_CANDRV_WRITE       103
#define IOCTL_CANDRV_NUMPORT       104
#define INIT_CAN_DEVICE               100

// Error definitions
#define STATUS_IRQ_REQUEST_ERROR	1
#define STATUS_REQUEST_IO_REGION_ERROR	2
#define STATUS_SUCCESS			0
#define STATUS_DEVICE_CONFIGURATION_ERROR	3
#define STATUS_INSUFFICIENT_RESOURCES		4
#define STATUS_INVALID_DEVICE_REQUEST		5
#define STATUS_NO_DATA_DETECTED			6
#define STATUS_NO_DATAGRAM_PRESENT		7



#define CANDRV_NAME				"CAN_LinuxDrv"

typedef union
{
  unsigned char  uc[4];
  unsigned long ul;
} ULCONV;

typedef union
{
  unsigned char  uc[2];
  unsigned short uw;
} UWCONV;



typedef struct
{
  unsigned char 	*baseIO;        
  unsigned short    portOffset;      
  unsigned long     bitRate; 
  unsigned long		irq;
} INIT_DATA;    

typedef struct
{
	unsigned char *message;
	unsigned short channel;

} READ_DATA;    

typedef struct
{
	unsigned char *message;
	unsigned short channel;

} WRITE_DATA;

typedef struct
{
	unsigned long version;
} VERSION_DATA;    

typedef struct
{
	unsigned long numPorts;
} NUM_PORTS;    




//****************************************************************************
// DEFINES

// sja1000 registers, only PELICAN mode 
//#define MODE                   0      // mode register
//#define COMMAND                1
//#define CHIPSTATUS             2
//#define INTERRUPT_STATUS       3
//#define INTERRUPT_ENABLE       4      // acceptance code
//#define TIMING0                6      // bus timing 0
//#define TIMING1                7      // bus timing 1
//#define OUTPUT_CONTROL         8      // output control
//#define TESTREG                9
//
//#define ARBIT_LOST_CAPTURE    11      // transmit buffer: Identifier
//#define ERROR_CODE_CAPTURE    12      // RTR bit und data length code
//#define ERROR_WARNING_LIMIT   13      // start byte of data field
//#define RX_ERROR_COUNTER      14
//#define TX_ERROR_COUNTER      15
//
//#define ACCEPTANCE_CODE_BASE  16
//#define RECEIVE_FRAME_BASE    16
//#define TRANSMIT_FRAME_BASE   16
//
//#define ACCEPTANCE_MASK_BASE  20
//
//#define RECEIVE_MSG_COUNTER   29
//#define RECEIVE_START_ADDRESS 30
//
//#define CLKDIVIDER            31      // set bit rate and pelican mode

// sja1000 registers, only BASICCAN mode 
#define CONTROL                0     
#define COMMAND                1
#define CHIPSTATUS             2
#define INTERRUPT_STATUS       3
#define ACCEPTANCE_CODE        4      // acceptance code
#define ACCEPTANCE_MASK        5      // acceptance mask
#define TIMING0                6      // bus timing 0
#define TIMING1                7      // bus timing 1
#define OUTPUT_CONTROL         8      // output control
#define TESTREG                9
#define TRANSMIT_FRAME_BASE   10
#define RECEIVE_FRAME_BASE    20
#define CLKDIVIDER            31      // set bit rate and pelican mode


// important sja1000 register contents, MODE register
//#define SLEEP_MODE             0x10
//#define ACCEPT_FILTER_MODE     0x08
//#define SELF_TEST_MODE         0x04
//#define LISTEN_ONLY_MODE       0x02
//#define RESET_MODE             0x01
//#define NORMAL_MODE            0x00

// important sja1000 register contents, CONTROL register
#define DATA_OVERRUN_INTERRUPT_ENABLE 0x10
#define BUS_ERROR_INTERRUPT_ENABLE    0x08
#define TRANSMIT_INTERRUPT_ENABLE     0x04
#define RECEIVE_INTERRUPT_ENABLE      0x02
#define RESET_MODE					  0x01


// COMMAND register
#define SLEEP_MODE             0x10
#define CLEAR_DATA_OVERRUN     0x08
#define RELEASE_RECEIVE_BUFFER 0x04
#define ABORT_TRANSMISSION     0x02
#define TRANSMISSION_REQUEST   0x01

// CHIPSTATUS register
#define BUS_STATUS             0x80
#define ERROR_STATUS           0x40
#define TRANSMIT_STATUS        0x20
#define RECEIVE_STATUS         0x10
#define TRANS_COMPLETE_STATUS  0x08
#define TRANS_BUFFER_STATUS    0x04
#define DATA_OVERRUN_STATUS    0x02
#define RECEIVE_BUFFER_STATUS  0x01

// OUTPUT CONTROL register
#define OUTPUT_CONTROL_TRANSISTOR_P1  0x80
#define OUTPUT_CONTROL_TRANSISTOR_N1  0x40
#define OUTPUT_CONTROL_POLARITY_1     0x20
#define OUTPUT_CONTROL_TRANSISTOR_P0  0x10
#define OUTPUT_CONTROL_TRANSISTOR_N0  0x08
#define OUTPUT_CONTROL_POLARITY_0     0x04
#define OUTPUT_CONTROL_MODE_1         0x02
#define OUTPUT_CONTROL_MODE_0         0x01


// INTERRUPT STATUS register
#define WAKE_UP_INTERRUPT      0x10
#define DATA_OVERRUN_INTERRUPT 0x08
#define ERROR_WARN_INTERRUPT   0x04
#define TRANSMIT_INTERRUPT     0x02
#define RECEIVE_INTERRUPT      0x01
#define INTERRUPT_MASK		   0x03
#define ERROR_PASSIV_INTERRUPT 	0x20
#define BUS_ERROR_INTERRUPT	0x80




// TRANSMIT or RECEIVE BUFFER
#define BUFFER_EFF                    0x80 // set for 29 bit identifier
#define BUFFER_RTR                    0x40 // set for RTR request
#define BUFFER_DLC_MASK               0x0f

// CLKDIVIDER register
#define CAN_MODE                      0x80
#define CAN_BYPASS                    0x40
#define RXINT_OUTPUT_ENABLE           0x20
#define CLOCK_OFF                     0x08
#define CLOCK_DIVIDER_MASK            0x07

// additional informations
#define CLOCK_HZ                  16000000 // crystal frequency

// time for mode register to change mode
#define MODE_REGISTER_SWITCH_TIME 100 // msec

// some CLKDIVIDER register contents, hardware architecture dependend
#define PELICAN_SINGLE  (CAN_MODE | CAN_BYPASS | 0x07 | CLOCK_OFF)
#define PELICAN_MASTER  (CAN_MODE | CAN_BYPASS | 0x07            )
#define PELICAN_DEFAULT (CAN_MODE                                )
#define CHIP_RESET      PELICAN_SINGLE

// some CLKDIVIDER register contents, hardware architecture dependend
#define BASICCAN_SINGLE  (CAN_BYPASS | 0x07 | CLOCK_OFF)
#define BASICCAN_MASTER  (CAN_BYPASS | 0x07            )
#define BASICCAN_DEFAULT (0)
#define BASICAN_CHIP_RESET      BASICCAN_SINGLE

// hardware depended setup for OUTPUT_CONTROL register
#define OUTPUT_CONTROL_SETUP (OUTPUT_CONTROL_TRANSISTOR_P0 | OUTPUT_CONTROL_TRANSISTOR_N0 | OUTPUT_CONTROL_MODE_1)

// the interrupt enables
#define INTERRUPT_ENABLE_SETUP (RECEIVE_INTERRUPT_ENABLE | TRANSMIT_INTERRUPT_ENABLE)

// the maximum number of handled messages in one interrupt
#define MAX_MESSAGES_PER_INTERRUPT 20

#ifndef PCAN_SJA1000_DONT_LOOP_ON_ISR
// the maximum number of handled sja1000 interrupts in 1 handler entry
#define MAX_INTERRUPTS_PER_ENTRY   200
#endif

#define MAX_TSEG1  15
#define MAX_TSEG2  7
#define BTR1_SAM   (1<<1)


#define __BIG_ENDIAN		1

#define READBUFFER_SIZE      80                            // buffers used in read and write call
#define WRITEBUFFER_SIZE     80

#define READ_MESSAGE_COUNT  500                            // max read message count
#define WRITE_MESSAGE_COUNT  50                            // max write message count

/* special address description flags for the CAN_ID */
#define CAN_EFF_FLAG 0x8000U /* EFF/SFF is set in the MSB */
#define CAN_RTR_FLAG 0x40000000U /* remote transmission request */
#define CAN_ERR_FLAG 0x20000000U /* error frame */

// Activity states
#define ACTIVITY_NONE        0          // LED off           - set when the channel is created or deleted
#define ACTIVITY_INITIALIZED 1          // LED on            - set when the channel is initialized
#define ACTIVITY_IDLE        2          // LED slow blinking - set when the channel is ready to receive or transmit
#define ACTIVITY_XMIT        3          // LED fast blinking - set when the channel has received or transmitted

/* error class (mask) in can_id */
#define CAN_ERR_TX_TIMEOUT   0x00000001U /* TX timeout (by netdevice driver) */
#define CAN_ERR_LOSTARB      0x00000002U /* lost arbitration    / data[0]    */
#define CAN_ERR_CRTL         0x00000004U /* controller problems / data[1]    */
#define CAN_ERR_PROT         0x00000008U /* protocol violations / data[2..3] */
#define CAN_ERR_TRX          0x00000010U /* transceiver status  / data[4]    */
#define CAN_ERR_ACK          0x00000020U /* received no ACK on transmission */
//#define CAN_ERR_BUSOFF       0x00000040U /* bus off */
#define CAN_ERR_BUSERROR     0x00000080U /* bus error (may flood!) */
#define CAN_ERR_RESTARTED    0x00000100U /* controller restarted */


#define CAN_ERR_DLC 8 /* dlc for error frames */


//****************************************************************************
// error codes
#define CAN_ERR_OK             0x0000  // no error
#define CAN_ERR_XMTFULL        0x0001  // transmit buffer full
#define CAN_ERR_OVERRUN        0x0002  // overrun in receive buffer
#define CAN_ERR_BUSLIGHT       0x0004  // bus error, errorcounter limit reached
#define CAN_ERR_BUSHEAVY       0x0008  // bus error, errorcounter limit reached
#define CAN_ERR_BUSOFF         0x0010  // bus error, 'bus off' state entered
#define CAN_ERR_QRCVEMPTY      0x0020  // receive queue is empty
#define CAN_ERR_QOVERRUN       0x0040  // receive queue overrun
#define CAN_ERR_QXMTFULL       0x0080  // transmit queue full 
#define CAN_ERR_REGTEST        0x0100  // test of controller registers failed
#define CAN_ERR_NOVXD          0x0200  // Win95/98/ME only
#define CAN_ERR_RESOURCE       0x2000  // can't create resource
#define CAN_ERR_ILLPARAMTYPE   0x4000  // illegal parameter
#define CAN_ERR_ILLPARAMVAL    0x8000  // value out of range
#define CAN_ERRMASK_ILLHANDLE  0x1C00  // wrong handle, handle error

#define CAN_ERR_BUSOFF_NETDEV CAN_ERR_BUSOFF


/* error status of CAN-controller / data[1] */
#define CAN_ERR_CRTL_UNSPEC      0x00 /* unspecified */
#define CAN_ERR_CRTL_RX_OVERFLOW 0x01 /* RX buffer overflow */
#define CAN_ERR_CRTL_TX_OVERFLOW 0x02 /* TX buffer overflow */
#define CAN_ERR_CRTL_RX_WARNING  0x04 /* reached warning level for RX errors */
#define CAN_ERR_CRTL_TX_WARNING  0x08 /* reached warning level for TX errors */
#define CAN_ERR_CRTL_RX_PASSIVE  0x10 /* reached error passive status RX */
#define CAN_ERR_CRTL_TX_PASSIVE  0x20 /* reached error passive status TX */
				      /* (at least one error counter exceeds */
				      /* the protocol-defined level of 127)  */

#define CAN_ERROR_ACTIVE     0          // CAN-Bus error states for busStatus - initial and normal state
#define CAN_ERROR_PASSIVE    1          // receive only state
#define CAN_BUS_OFF          2          // switched off from Bus

#define CHANNEL_SINGLE 0                                   // this is a single channel device
#define CHANNEL_MASTER 1                                   // multi channel device, this device is master
#define CHANNEL_SLAVE  2                                   // multi channel device, this is slave

// parameter wBTR0BTR1
// bitrate codes of BTR0/BTR1 registers
#define CAN_BAUD_1M     0x0014                             //   1 MBit/s
#define CAN_BAUD_500K   0x001C                             // 500 kBit/s
#define CAN_BAUD_250K   0x011C                             // 250 kBit/s
#define CAN_BAUD_125K   0x031C                             // 125 kBit/s
#define CAN_BAUD_100K   0x432F                             // 100 kBit/s
#define CAN_BAUD_50K    0x472F                             //  50 kBit/s
#define CAN_BAUD_20K    0x532F                             //  20 kBit/s
#define CAN_BAUD_10K    0x672F                             //  10 kBit/s
#define CAN_BAUD_5K     0x7F7F                             //   5 kBit/s

//****************************************************************************
// MSGTYPE bits of element MSGTYPE in structure TPCANMsg
#define MSGTYPE_STATUS        0x80     // used to mark a status TPCANMsg
#define MSGTYPE_EXTENDED      0x02     // declares a extended frame
#define MSGTYPE_RTR           0x01     // marks a remote frame
#define MSGTYPE_STANDARD      0x00     // marks a standard frame

/* valid bits in CAN ID for frame formats */
#define CAN_SFF_MASK 0x000007FFU /* standard frame format (SFF) */
#define CAN_EFF_MASK 0x1FFFFFFFU /* extended frame format (EFF) */
#define CAN_ERR_MASK 0x1FFFFFFFU /* omit EFF, RTR, ERR flags */


// ----------------------------------------------------------------------------
// Software related defines and equates
// ----------------------------------------------------------------------------


typedef struct
{
	unsigned short        wStepSize;                                    // size of bytes to step to next entry
	unsigned short        wCopySize;                                    // size of bytes to copy
	void       *bufferBegin;                                 // points to first element
	void       *bufferEnd;                                   // points to the last element
	unsigned long        nCount;                                       // max count of elements in fifo
	unsigned long        nStored;                                      // count of currently received and stored messages
	unsigned long        dwTotal;                                      // received messages
	void       *r;                                           // points to the next Msg to read into the read buffer
	void       *w;                                           // points to the next Msg to write into read buffer
	//spinlock_t lock;                                         // mutual exclusion lock
} FIFO_MANAGER;



typedef struct 
{
  unsigned long ID;              // 11/29 bit code
  unsigned char  MSGTYPE;         // bits of MSGTYPE_*
  unsigned char  LEN;             // count of data bytes (0..8)
  unsigned char  DATA[8];         // data bytes, up to 8
} TPCANMsg;              // for PCAN_WRITE_MSG


typedef struct
{
  TPCANMsg Msg;          // the above message
  unsigned long    dwTime;       // a timestamp in msec, read only
  unsigned short     wUsec;        // remainder in micro-seconds
} TPCANRdMsg;    

struct can_frame {
	unsigned short   can_id;  
	unsigned char    data[8]; //__attribute__ ((aligned(8)));
};



typedef struct channel                // this structure holds various channel properties
{
    unsigned char *PortBase;										// I/O port base address for channel
	FIFO_MANAGER readFifo;                                   // manages the read fifo
	FIFO_MANAGER writeFifo;                                  // manages the write fifo
	TPCANRdMsg   rMsg[READ_MESSAGE_COUNT];                   // all read messages
	TPCANMsg     wMsg[WRITE_MESSAGE_COUNT];                  // all write messages
	void *filter;											 // a ID filter - currently associated to device
} CHAN;

typedef struct
{
  unsigned char 	*baseIO;        
  unsigned short    portOffset;      
  unsigned long     bitRate; 
  unsigned long		irq;
  CHAN chan[2];
  unsigned long numChannels;
} DRIVER_DATA;

int set_reset_mode(DRIVER_DATA *pDriver, unsigned short channel);
int set_normal_mode(DRIVER_DATA *pDriver, unsigned short channel);
void device_irq_enable(DRIVER_DATA *pDriver);
void device_irq_disable(DRIVER_DATA *pDriver);
int device_read_frames(DRIVER_DATA *pDriver, unsigned short channel);
int device_write_frame(DRIVER_DATA *pDriver, struct can_frame *cf, unsigned short channel);
int isr_device_write(DRIVER_DATA *pDriver, unsigned short channel);
int device_write(DRIVER_DATA *pDriver, unsigned short channel);
int device_baud_rate(int rate, int flags);
unsigned short device_bitrate(unsigned long dwBitRate);
int pcan_chardev_rx(DRIVER_DATA *pDriver, struct can_frame *cf, unsigned short channel);
void guarded_write_command(DRIVER_DATA *pDriver, unsigned char data, unsigned short channel);
void frame2msg(struct can_frame *cf, TPCANMsg *msg);
void msg2frame(struct can_frame *cf, TPCANMsg *msg);
int pcan_fifo_get(FIFO_MANAGER *anchor, void *pvGetData);
int pcan_fifo_put(FIFO_MANAGER *anchor, void *pvPutData);
int pcan_fifo_init(FIFO_MANAGER *anchor, void *bufferBegin, void *bufferEnd, int nCount, unsigned short wCopySize);
int pcan_fifo_reset(FIFO_MANAGER *anchor);

#endif

